// Author: A.Voss@FH-Aachen.de
//
// variadic templates
// tags: template

#include <iostream>
#include <cstdlib>
using namespace std;

#include "../../include/tools.h"

template <typename T>
void my_printf(T t)
{
    cout << t;
}

template<typename T0, typename... TN>
void my_printf(T0 t0, TN... tn)       // TN pack expansion
{
    cout << t0 << " ; ";
    my_printf(tn...);
}

template <typename... Args> 
struct NumberOfArgs {
    static const int value = sizeof...(Args);
};

template <typename... Args>
class MyTuple;

template <>
class MyTuple<>
{};

template <typename T0, typename... TN>
class MyTuple<T0,TN...> : private MyTuple<TN...>
{
    T0 data;
public:
    MyTuple(const T0 & t0, const TN & ...tn) 
        : MyTuple<TN...>(tn...), data(t0)
    {}

    T0 const& t0() const { return data; }

    const MyTuple<TN...> & tn() const { return *this; }
};


int main()
{
    tools_log(); cout << "use of variadic templates" << endl;
    
    tools_log(); cout << "  a set: { "; my_printf("text",12.34,-23,'X'); 
                 cout << " }" << endl;
    
    tools_log(); cout << "  number of tepl.args: " << NumberOfArgs<int,double,char>::value << endl;
    
    MyTuple<int,double> e(12,56.7);
    tools_log(); cout << "  first: " << e.t0() << ", second: " << e.tn().t0() << endl;

    return EXIT_SUCCESS;
}

