// Author: A.Voss@FH-Aachen.de
//
// user literals
// tags: template, literals

#include <iostream>
#include <cstdlib>
using namespace std;

#include "../../include/tools.h"

// helper class to provide different templates and units
struct DistanceUnit {
    static const char * unit() { return "m"; }
};
struct TimeUnit {
    static const char * unit() { return "s"; }
};
struct SpeedUnit {
    static const char * unit() { return "m/s"; }
};

// a simple template providing a double value and an unit
template <typename T>
class Value {
public:
    typedef T unit_type;
    double _value;
    
    Value(double v) : _value(v) { }
    
    operator double() const { return _value; }
    
    friend ostream & operator<<(ostream & os, const Value & v) {
        os << v._value << unit_type::unit();
        return os;
    }
};

// equivalent to typedefs, create a specific type 
using Distance = Value<DistanceUnit>;
using Time = Value<TimeUnit>;
using Speed = Value<SpeedUnit>;

// and now the suffixes on doubles (!) note: 12_m would not compile
Time operator"" _s(long double s) { return Time((double)s); }
Distance operator"" _m(long double s) { return Distance((double)s); }

// and an operator /
Speed operator/(const Distance & d, const Time & t) { return Speed((double)d/(double)t); }

int main()
{
    tools_log(); cout << "use of user literals" << endl;

    // define a distance, a time and a speed-var
    auto s = 120.0_m;
    auto t = 10.0_s;
    auto v = s/t;
    
    tools_log(); cout << "  dist: " << s << ", time: " << t << ", speed: " << v << endl;
    
    // or explicitly in type
    Distance ac_k = 70.7_m*1000;
    tools_log(); cout << "  dist ac-k: " << ac_k << ", or, as double: " << (double)ac_k << endl;

    return EXIT_SUCCESS;
}

